# -*- coding: utf-8 -*-

"""
Tests for ambiguity checking and reporting.
"""

from graphtransliterator import (
    AmbiguousTransliterationRulesException,
    GraphTransliterator, TransliterationRule
)
from graphtransliterator.ambiguity import _easyreading_rule
import pytest


def test_GraphParser_check_ambiguity():
    """ Test for rules that can both match the same thing."""

    yaml_for_test = r"""
        tokens:
          a: [token, class1, class2]
          b: [token, class1, class2]
          ' ': [wb]
        rules:
          a <class1>: A<class1> # these should be ambiguous
          a <class2>: A<class2>

          <class1> a: <class1>A  # these should be ambiguous
          <class2> a: <class2>A # these should be ambiguous

          (<class1> b) a (b <class2>): A # ambigous
          (<class2> b) a (b <class1>): A # ambiguous
          a: A # not ambiguous
        whitespace:
          default: ' '
          token_class: 'wb'
          consolidate: true
        """
    with pytest.raises(AmbiguousTransliterationRulesException):
        GraphTransliterator.from_yaml(yaml_for_test, check_for_ambiguity=True)
    # check that ambiguity matches if rules are of different shape
    yaml = """
        tokens:
          a: []
          ' ': [wb]
        rules:
          <wb> a: _A
          a <wb>: A_
          a: a
          ' ': ' '
        whitespace:
          default: " "        # default whitespace token
          consolidate: true  # whitespace should be consolidated
          token_class: wb     # whitespace token class
        """
    with pytest.raises(AmbiguousTransliterationRulesException):
        GraphTransliterator.from_yaml(yaml, check_for_ambiguity=True)


def test_GraphTransliterator_easy_reading():
    assert (
        _easyreading_rule(
            TransliterationRule("", ["class_a"], [], ["a"], [], ["class_a"], 0)
        )
        == "<class_a> a <class_a>"
    )
    assert (
        _easyreading_rule(
            TransliterationRule("", ["class_a"], [], ["a"], [], ["class_a"], 0)
        )
        == "<class_a> a <class_a>"
    )
    assert (
        _easyreading_rule(TransliterationRule("", [], ["b"], ["a"], ["b"], [], 0))
        == "(b) a (b)"
    )
    assert (
        _easyreading_rule(
            TransliterationRule("", ["class_a"], ["b"], ["a"], ["b"], ["class_a"], 0)
        )
        == "(<class_a> b) a (b <class_a>)"
    )



# -*- coding: utf-8 -*-
"""
Functions used to compress and decompress a GraphTransliterator.
"""
import math


def compress_config(config, compression_level=1):
    """
    Compress configuration to minimize JSON size.

    Compression uses integer lookup for classes, tokens, and graph node types, as well
    as numerous tuples. It contains the graph definition but not easily generatable
    items.


    Parameters
    ----------
    config : `dict`
        Configuration of GraphTransliterator, e.g. from `dump` or `loads`.

    Returns
    -------
    `tuple`
        Compressed configuration consisting of:

            Class names : `tuple` of `str`
                Provides id of classes when compressed to integer index

            Token strings : (`tuple` of `str`)
                Provides id of tokens when compressed to integer index

            Token classes : (`tuple` of `int`) or 0 (`int`) for None

            Production rules : (`tuple` of `tuple`)

                Production: `str`

                Previous class ids : (`tuple` of `int`) or 0 (`int`) for None

                Previous token ids : (`tuple` of `int`) or 0 (`int`) for None

                Token ids : (`tuple` of `int`)

                Next token ids : (`tuple` of `int`) or 0 (`int`) for None

                Next class ids : (`tuple` of `int`) or 0 (`int`) for None

            Cost : Negative integer cost (`int`)

            Graph : `tuple`

                Node types : `tuple` of `str`
                    Provides id of node types when compressed to integer index

                Nodes : `tuple` of `tuple`
                    All node types begin with the following fields:

                        Node type id : `int`
                        Accepting: 0 or 1 (`int`) for False or True

                    `Start` nodes additionally contain:

                        Ordered Children : `dict` of token to node ids {`int`: `int`}
                            The rules key (`__rules__`) is -1

                    `rule` nodes additionally contain:
                        Rule key: `int`

                    `token` nodes additionally contain:

                        Token id: `int`

                        Ordered Children: `dict` of token to node ids {`int`: `int`}

                            The rules key (`__rules__`) is -1

                Edges: `dict` of {`int`: `dict` of {`int`: `tuple`}}

                    Dictionary points from head id to tail id to edge data. The data
                    consists of the following:

                        Constraints: `tuple` or `0` (`int`) for None

                            Previous classes : class ids (`list` of `int`), or 0 (`int`)

                            Previous tokens : token ids (`list` of `int`), or 0 (`int`)

                            Next tokens: token ids (`list` of `int`), or 0 (`int`)

                            Next classes : class ids (`list` of `int`), or 0 (`int`)

                        Cost : Negative integer cost (`int`)

                        Token: Token id (`int`) or -1 for None
    """

    # No compression, human-readable
    if compression_level == 0:
        return config

    def compressed_cost(x):
        return -1 * round((1 / (2 ** x - 1) - 1))  # convert to -int

    def compress_edge_data(data):

        constraints = data.get("constraints")

        def _class_ids_of(constraint_attr):
            v = constraints.get(constraint_attr)
            if not v:
                return 0
            return [_class_id[_] for _ in v]

        def _token_ids_of(constraint_attr):
            v = constraints.get(constraint_attr)
            if not v:
                return 0
            return [_token_id[_] for _ in v]

        new_constraints = 0

        if constraints:
            new_constraints = [
                _class_ids_of("prev_classes"),
                _token_ids_of("prev_tokens"),
                _token_ids_of("next_tokens"),
                _class_ids_of("next_classes"),
            ]
        new_cost = compressed_cost(data.get("cost"))
        new_token = -1
        _token = data.get("token")
        if _token:
            new_token = _token_id[_token]

        return tuple([new_constraints, new_cost, new_token])

    def compress_node(_node):
        def compressed_ordered_children():
            x = _node.get("ordered_children")
            out = {}
            for k, v in x.items():
                if k == "__rules__":
                    out[-1] = v
                else:
                    out[_token_id[k]] = v
            return out

        _type_id = _nodetype_id[_node["type"]]
        _accepting = 1 if _node.get("accepting") else 0

        if _node["type"] == "Start":
            new_node = tuple([_type_id, _accepting, compressed_ordered_children()])
        elif _node["type"] == "rule":
            new_node = tuple([_type_id, _accepting, _node["rule_key"]])
        elif _node["type"] == "token":
            new_node = tuple(
                [
                    _type_id,
                    _accepting,
                    _token_id[_node["token"]],
                    compressed_ordered_children(),
                ]
            )
        return new_node

    token_list = tuple(sorted(config["tokens"].keys()))
    _token_id = {_: i for i, _ in enumerate(token_list)}

    class_list = tuple(sorted(set().union(*config["tokens"].values())))
    _class_id = {_: i for i, _ in enumerate(class_list)}
    tokens = tuple(
        tuple(_class_id[_] for _ in config["tokens"][tkn]) for tkn in token_list
    )
    nodetype_list = tuple(sorted(set([_["type"] for _ in config["graph"]["node"]])))
    _nodetype_id = {_: i for i, _ in enumerate(nodetype_list)}
    rules = tuple(
        (
            r["production"],
            tuple(_class_id[_] for _ in r.get("prev_classes"))
            if r.get("prev_classes")
            else 0,
            tuple(_token_id[_] for _ in r.get("prev_tokens"))
            if r.get("prev_tokens")
            else 0,
            tuple(_token_id[_] for _ in r.get("tokens")),
            tuple(_token_id[_] for _ in r.get("next_tokens"))
            if r.get("next_tokens")
            else 0,
            tuple(_class_id[_] for _ in r.get("next_classes"))
            if r.get("next_classes")
            else 0,
            compressed_cost(r["cost"]),
        )
        for r in config["rules"]
    )
    whitespace = tuple(
        [
            config["whitespace"]["default"],
            config["whitespace"]["token_class"],
            1 if config["whitespace"]["consolidate"] else 0,
        ]
    )
    onmatch_rules = (
        tuple(
            tuple(
                [
                    tuple(_class_id[_] for _ in r["prev_classes"]),
                    tuple(_class_id[_] for _ in r["next_classes"]),
                    r["production"],
                ]
            )
            for r in config["onmatch_rules"]
        )
        if config.get("onmatch_rules")
        else 0
    )
    metadata = config.get("metadata", 0)

    if compression_level == 1:
        # Compress with generated graph; no information loss.
        _graph = config.get("graph")
        node = tuple(compress_node(_) for _ in _graph["node"])
        _edge = _graph["edge"]
        edge = {
            int(head_id): {
                int(tail_id): compress_edge_data(edge_data)
                for tail_id, edge_data in tail.items()
            }
            for head_id, tail in _edge.items()
        }

        return tuple(
            [
                class_list,
                token_list,
                tokens,
                rules,
                whitespace,
                onmatch_rules,
                metadata,
                [nodetype_list, node, edge],
            ]
        )
    elif compression_level == 2:
        # Compress without graph; no information loss.
        return tuple(
            [
                class_list,
                token_list,
                tokens,
                rules,
                whitespace,
                onmatch_rules,
                metadata,
                None,
            ]
        )


def strip_empty(d):
    """Strips entries of dict with no value, but allow zero."""
    return {
        k: v
        for k, v in d.items()
        if v or (type(v) == int and v == 0) or (type(v) == str and v == "")
    }


def decompress_config(compressed_config):
    def decompressed_cost(x):
        """x will be negative."""
        return math.log2(1 + 1 / (1 - x))

    def decompress_node(_node):
        def decompressed_ordered_children(index):
            x = _node[index]
            out = {}
            for k, v in x.items():
                k = int(k)
                if k == -1:
                    out["__rules__"] = v
                else:
                    out[_token_from_id[k]] = v
            return out

        node_type = _nodetype_from_id[_node[0]]
        accepting = True if _node[1] else False
        if node_type == "Start":
            new_node = {
                "type": node_type,
                "accepting": accepting,
                "ordered_children": decompressed_ordered_children(2),
            }
        elif node_type == "rule":
            new_node = {"type": node_type, "accepting": accepting, "rule_key": _node[2]}
        elif node_type == "token":
            new_node = {
                "type": node_type,
                "accepting": accepting,
                "token": _token_from_id[_node[2]],
                "ordered_children": decompressed_ordered_children(3),
            }
        return strip_empty(new_node)

    def decompress_edge_data(data):

        [_constraints, _cost, _token] = data

        out = {}

        def _class_from_ids_of(index):
            v = _constraints[index]
            if v == 0:
                return None
            return [_class_from_id[_] for _ in v]

        def _token_from_ids_of(index):
            v = _constraints[index]
            if v == 0:
                return None
            return [_token_from_id[_] for _ in v]

        if _constraints:
            # filter out unused values
            out["constraints"] = strip_empty(
                {
                    "prev_classes": _class_from_ids_of(0),
                    "prev_tokens": _token_from_ids_of(1),
                    "next_tokens": _token_from_ids_of(2),
                    "next_classes": _class_from_ids_of(3),
                }
            )

        out["cost"] = decompressed_cost(_cost)

        if _token != -1:  # -1 indicates no token
            out["token"] = _token_from_id[_token]

        return out

    [
        _class_list,
        _token_list,
        _tokens,
        _rules,
        _whitespace,
        _onmatch_rules,
        metadata,
        _graph,
    ] = compressed_config

    _token_list = list(_token_list)
    _token_from_id = {i: _ for i, _ in enumerate(_token_list)}
    _class_list = list(_class_list)
    _class_from_id = {i: _ for i, _ in enumerate(_class_list)}
    tokens = {
        tkn: [_class_list[_] for _ in _tokens[i]] for i, tkn in enumerate(_token_list)
    }
    rules = [
        strip_empty(
            {
                "production": _[0],
                "prev_classes": [_class_list[i] for i in _[1]] if _[1] else [],
                "prev_tokens": [_token_list[i] for i in _[2]] if _[2] else [],
                "tokens": [_token_list[i] for i in _[3]],
                "next_tokens": [_token_list[i] for i in _[4]] if _[4] else [],
                "next_classes": [_class_list[i] for i in _[5]] if _[5] else [],
                "cost": decompressed_cost(_[6]),
            }
        )
        for _ in _rules
    ]
    whitespace = {
        "default": _whitespace[0],
        "token_class": _whitespace[1],
        "consolidate": _whitespace[2],
    }
    onmatch_rules = (
        [
            {
                "prev_classes": [_class_from_id[_] for _ in r[0]],
                "next_classes": [_class_from_id[_] for _ in r[1]],
                "production": r[2],
            }
            for r in _onmatch_rules
        ]
        if _onmatch_rules
        else None
    )
    if not _graph:  # no graph due to compression
        graph = None
    else:
        [_nodetype_list, _nodes, _edges] = _graph
        _nodetype_from_id = {i: _ for i, _ in enumerate(_nodetype_list)}
        node = [decompress_node(_) for _ in _nodes]
        edge = {
            int(head_id): {
                int(tail_id): decompress_edge_data(edge_data)
                for tail_id, edge_data in tail.items()
            }
            for head_id, tail in _edges.items()
        }
        graph = {"node": node, "edge": edge}
    return {
        "tokens": tokens,
        "rules": rules,
        "whitespace": whitespace,
        "onmatch_rules": onmatch_rules,
        "graph": graph,
        "metadata": metadata,
    }



# -*- coding: utf-8 -*-

"""
GraphTransliterator ambiguity-checking functions.
"""
from .exceptions import AmbiguousTransliterationRulesException
import itertools
import logging


def check_for_ambiguity(transliterator):
    """
    Check if multiple transliteration rules could match the same tokens.

    This function first groups the transliteration rules by number of
    tokens. It then checks to see if any pair of the same cost would match
    the same sequence of tokens. If so, it finally checks if a less costly
    rule would match those particular sequences. If not, there is
    ambiguity.

    Details of all ambiguity are sent in a :func:`logging.warning`.

    Note
    ----
    Called during initialization if ``check_ambiguity`` is set.

    Raises
    ------
    AmbiguousTransliterationRulesException
        Multiple transliteration rules could match the same tokens.

    Example
    -------
    .. jupyter-execute::

      yaml_filename = '''
      tokens:
        a: [class1, class2]
        ' ': [wb]
      rules:
        <class1> a: AW
        <class2> a: AA # ambiguous rule
      whitespace:
        default: ' '
        consolidate: True
        token_class: wb
      '''
      gt = GraphTransliterator.from_yaml(yaml_, check_ambiguity=False)
      gt.check_for_ambiguity()

    """
    ambiguity = False

    all_tokens = set(transliterator._tokens.keys())

    rules = transliterator._rules

    if not rules:
        return True

    max_prev = [_count_of_prev(rule) for rule in rules]
    global_max_prev = max(max_prev)
    max_curr_next = [_count_of_curr_and_next(rule) for rule in rules]
    global_max_curr_next = max(max_curr_next)

    # Generate a matrix of rules, where width is the max of
    # any previous tokens/classes + max of current/next tokens/classes.
    # Each rule's specifications starting from the max of the previous
    # tokens/classes. Other positions are filled by the set of all possible
    # tokens.

    matrix = []

    width = global_max_prev + global_max_curr_next

    for i, rule in enumerate(rules):
        row = [all_tokens] * (global_max_prev - max_prev[i])
        row += _tokens_possible(rule, transliterator._tokens_by_class)
        row += [all_tokens] * (width - len(row))
        matrix += [row]

    def full_intersection(i, j):
        """ Intersection of  matrix[i] and matrix[j], else None."""

        intersections = []
        for k in range(width):
            intersection = matrix[i][k].intersection(matrix[j][k])
            if not intersection:
                return None
            intersections += [intersection]
        return intersections

    def covered_by(intersection, row):
        """Check if intersection is covered by row."""
        for i in range(len(intersection)):
            diff = intersection[i].difference(row[i])
            if diff:
                return False
        return True

    # Iterate through rules based on cost (number of tokens). If there are
    # ambiguities, then see if a less costly rule would match the rule. If it does
    # not, there is ambiguity.

    for _group_val, group_iter in itertools.groupby(
        enumerate(transliterator._rules), key=lambda x: x[1].cost
    ):

        group = list(group_iter)
        if len(group) == 1:
            continue
        for i in range(len(group) - 1):
            for j in range(i + 1, len(group)):
                i_index = group[i][0]
                j_index = group[j][0]
                intersection = full_intersection(i_index, j_index)
                if not intersection:
                    break

                # Check if a less costly rule matches intersection

                def covered_by_less_costly():
                    for r_i, rule in enumerate(rules):
                        if r_i in (i_index, j_index):
                            continue
                        if rule.cost > rules[i_index].cost:
                            continue
                        rule_tokens = matrix[r_i]
                        if covered_by(intersection, rule_tokens):
                            return True
                    return False

                if not covered_by_less_costly():
                    logging.warning(
                        "The pattern {} can be matched by both:\n"
                        "  {}\n"
                        "  {}\n".format(
                            intersection,
                            _easyreading_rule(rules[i_index]),
                            _easyreading_rule(rules[j_index]),
                        )
                    )
                    ambiguity = True
    if ambiguity:
        raise AmbiguousTransliterationRulesException


def _easyreading_rule(rule):
    """Get an easy-reading string of a rule."""

    def _token_str(x):
        return " ".join(x)

    def _class_str(x):
        return " ".join(["<%s>" % _ for _ in x])

    out = ""
    if rule.prev_classes and rule.prev_tokens:
        out = "({} {}) ".format(
            _class_str(rule.prev_classes), _token_str(rule.prev_tokens)
        )
    elif rule.prev_classes:
        out = "{} ".format(_class_str(rule.prev_classes))
    elif rule.prev_tokens:
        out = "({}) ".format(_token_str(rule.prev_tokens))

    out += _token_str(rule.tokens)

    if rule.next_tokens and rule.next_classes:
        out += " ({} {})".format(
            _token_str(rule.next_tokens), _class_str(rule.next_classes)
        )
    elif rule.next_tokens:
        out += " ({})".format(_token_str(rule.next_tokens))
    elif rule.next_classes:
        out += " {}".format(_class_str(rule.next_classes))
    return out


def _count_of_prev(rule):
    """Count previous tokens to be present before a match in a rule."""

    return len(rule.prev_classes or []) + len(rule.prev_tokens or [])


def _count_of_curr_and_next(rule):
    """Count tokens to be matched and those to follow them in rule."""

    return len(rule.tokens) + len(rule.next_tokens or []) + len(rule.next_classes or [])


def _prev_tokens_possible(rule, tokens_by_class):
    """`list` of set of possible preceding tokens for a rule."""

    return [tokens_by_class[_] for _ in rule.prev_classes or []] + [
        set([_]) for _ in rule.prev_tokens or []
    ]


def _curr_and_next_tokens_possible(rule, tokens_by_class):
    """`list` of sets of possible current and following tokens for a rule."""

    return (
        [set([_]) for _ in rule.tokens]
        + [set([_]) for _ in rule.next_tokens or []]
        + [tokens_by_class[_] for _ in rule.next_classes or []]
    )


def _tokens_possible(row, tokens_by_class):
    """`list` of sets of possible tokens matched for a rule."""

    return _prev_tokens_possible(row, tokens_by_class) + _curr_and_next_tokens_possible(
        row, tokens_by_class
    )



# -*- coding: utf-8 -*-

from graphtransliterator.exceptions import (
    IncompleteGraphCoverageException,
    IncompleteOnMatchRulesCoverageException,
)
import graphtransliterator.transliterators as transliterators
import pytest


def test_CoverageTransliterator():
    """Test coverage using transliterators.Example"""
    # tokens:
    #   a: [vowel]
    #   ' ': [whitespace]
    #   b: [consonant]
    # rules:
    #   a: A
    #   b: B
    #   ' ': ' '
    #   (<consonant> a) b (a <consonant>):  "!B!"
    # -----
    covt = transliterators.Example(coverage=True)
    # unvisited should raise error
    with pytest.raises(IncompleteGraphCoverageException):
        covt.check_coverage()
    # visiting only some nodes should raise error
    with pytest.raises(IncompleteGraphCoverageException):
        assert covt.transliterate("a") == "A"
        assert covt.transliterate("b") == "B"
        covt.check_coverage()
    # visiting all nodes but no onmatch rules should raise error
    with pytest.raises(IncompleteOnMatchRulesCoverageException):
        assert covt.transliterate("a ") == "A "
        assert covt.transliterate("b") == "B"
        assert covt.transliterate("babab") == "BA!B!AB"
        covt.check_coverage()
    # visiting everything should not raise an error
    assert covt.transliterate("aa ") == "A,A "
    assert covt.check_coverage()
    # clearing visited should raise an error when checking coverage
    with pytest.raises(IncompleteGraphCoverageException):
        covt.clear_visited()
        covt.check_coverage()



# -*- coding: utf-8 -*-

"""Console script for graphtransliterator."""

from graphtransliterator import (
    DEFAULT_COMPRESSION_LEVEL, HIGHEST_COMPRESSION_LEVEL, GraphTransliterator
)
import click
import json
import os
import re
import sys


@click.group()
@click.version_option()
def main():
    pass


def load_transliterator(source, **kwargs):
    """Loads transliterator (format, parameter)."""
    format, parameter = source
    if format == "bundled":
        mod = __import__("graphtransliterator.transliterators")
        transliterators_mod = mod.transliterators
        transliterator_class = getattr(transliterators_mod, parameter)
        return transliterator_class(**kwargs)
    elif format == "json":
        return GraphTransliterator.loads(parameter, **kwargs)
    elif format == "json_file":
        with open(parameter, "r") as f:
            return GraphTransliterator.loads(f.read(), **kwargs)
    elif format == "yaml_file":
        return GraphTransliterator.from_yaml_file(parameter, **kwargs)


@click.command()
@click.option(
    "--from",
    "-f",
    "from_",
    type=(click.Choice(["bundled", "json", "json_file", "yaml_file"]), str),
    help=(
        "Format (bundled/json/json_file/yaml_file) and "
        "source (name, JSON, or filename) of transliterator"
    ),
    nargs=2,
    required=True,
)
@click.option(
    "--to",
    "-t",
    type=click.Choice(["json", "python"]),
    help="Output format",
    nargs=1,
    required=False,
    show_default=True,
    default="python",
)
@click.option(
    "--check-ambiguity/--no-check-ambiguity",
    "-ca/-nca",
    default=False,
    help="Check for ambiguity.",
    show_default=True,
)
@click.option(
    "--ignore-errors/--no-ignore-errors",
    "-ie",
    "-nie",
    help="Ignore errors.",
    show_default=True,
    default=False,
)
@click.argument("input", nargs=-1)
def transliterate(from_, to, input, check_ambiguity, ignore_errors):
    """Transliterate INPUT."""
    transliterator = load_transliterator(
        from_, check_ambiguity=check_ambiguity, ignore_errors=ignore_errors
    )
    if len(input) == 1:
        output = transliterator.transliterate(input[0])
    else:
        output = [transliterator.transliterate(_) for _ in input]
    if to == "json":
        click.echo(json.dumps(output))
    else:
        click.echo(output)


@click.command()
@click.option(
    "--from",
    "-f",
    "from_",
    type=(click.Choice(["bundled", "yaml_file"]), str),
    help="Format (bundled/yaml_file) and source (name or filename) of transliterator",
    nargs=2,
    required=True,
)
@click.option(
    "--check-ambiguity/--no-check-ambiguity",
    "-ca/-nca",
    default=False,
    help="Check for ambiguity.",
    show_default=True,
)
@click.option(
    "--compression-level",
    "-cl",
    default=DEFAULT_COMPRESSION_LEVEL,
    help=f"Compression level, from 0 to {HIGHEST_COMPRESSION_LEVEL}",
    show_default=True,
)
def dump(from_, check_ambiguity, compression_level):
    """Dump transliterator as JSON."""
    transliterator = load_transliterator(from_, check_ambiguity=check_ambiguity)
    click.echo(transliterator.dumps(compression_level=compression_level))


@click.command()
@click.option(
    "--from",
    "-f",
    "from_",
    type=(click.Choice(["bundled", "json", "json_file", "yaml_file"]), str),
    help=(
        "Format (bundled/json/json_file/yaml_file) and "
        "source (name, JSON, or filename) of transliterator"
    ),
    nargs=2,
    required=True,
)
@click.option(
    "--check-ambiguity/--no-check-ambiguity",
    "-ca/-nca",
    default=False,
    help="Check for ambiguity.",
    show_default=True,
)
def generate_tests(from_, check_ambiguity):
    """Generate tests as YAML."""
    import graphtransliterator.transliterators  # pragma: no cover
    transliterator = load_transliterator(from_, check_ambiguity=check_ambiguity)
    yaml_tests = graphtransliterator.transliterators.Bundled.generate_yaml_tests(
        transliterator
    )
    click.echo(yaml_tests)


@click.command()
@click.option(
    "--check-ambiguity/--no-check-ambiguity",
    "-ca/-nca",
    default=False,
    help="Check for ambiguity.",
    show_default=True,
)
@click.argument("bundled", nargs=1)
def test(bundled, check_ambiguity):
    """Test BUNDLED transliterator."""
    transliterator = load_transliterator(
        ["bundled", bundled], check_ambiguity=check_ambiguity
    )
    click.echo(transliterator.run_yaml_tests())


@click.command()
def list_bundled():
    """List BUNDLED transliterators."""
    import graphtransliterator.transliterators as transliterators
    click.echo("Bundled transliterators:")
    for _ in transliterators.iter_names():
        click.echo(f"  {_}")


@click.argument("bundled", nargs=1)
@click.option('--regex', '-re', is_flag=True,
              help="Match transliterators using regular expression.")
@click.command()
def make_json(bundled, regex):
    """Make JSON rules of BUNDLED transliterator(s)."""
    import graphtransliterator.transliterators as transliterators  # pragma: no cover
    if not regex:
        bundled = f"^{bundled}$"
    to_dump = [_ for _ in transliterators.iter_names() if re.match(bundled, _)]
    if not to_dump:
        click.echo(f"No bundled transliterator found matching /{bundled}/.")
        click.echo(
            'Try "graphtransliterator list-bundled" for a list.'
        )
        return
    for _ in to_dump:
        transliterator_class = getattr(transliterators, _)
        transliterator = transliterator_class.new(method="yaml")
        json_filename = os.path.join(
            transliterator.directory, transliterator.name + ".json"
        )
        with open(json_filename, "w") as f:
            f.write(transliterator.dumps())
        click.echo(f"Made JSON of {_}.")


main.add_command(dump)
main.add_command(generate_tests)
main.add_command(list_bundled)
main.add_command(make_json)
main.add_command(test)
main.add_command(transliterate)

if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



# -*- coding: utf-8 -*-

# flake8: noqa

"""
graphtransliterator
~~~~~~~~~~~~~~~~~~
Graph Transliterator is a Python module enabling rule-based transliteration
of the tokens of an input string to an output string. It can be used to
transliterate the symbols of one language into those of another, or to
convert between transliteration systems.

:copyright: © 2019 Michigan State University.
:license: MIT, see LICENSE.rst for more details.
"""

__author__ = """A. Sean Pue"""
__email__ = "pue@msu.edu"
__version__ = "1.1.1"

# Core classes
from .core import CoverageTransliterator, GraphTransliterator, GraphTransliteratorSchema

# Constants
from .core import DEFAULT_COMPRESSION_LEVEL, HIGHEST_COMPRESSION_LEVEL

# Exceptions
from .exceptions import (
    GraphTransliteratorException,
    AmbiguousTransliterationRulesException,
    NoMatchingTransliterationRuleException,
    UnrecognizableInputTokenException,
)

# Graphs
from .graphs import (
    DirectedGraph,
    VisitLoggingDirectedGraph,
    VisitLoggingDict,
    VisitLoggingList,
)

# Rules
from .rules import TransliterationRule, OnMatchRule, WhitespaceRules

# Schemas
from .schemas import (
    DirectedGraphSchema,
    EasyReadingSettingsSchema,
    OnMatchRuleSchema,
    SettingsSchema,
    TransliterationRuleSchema,
    WhitespaceDictSettingsSchema,
    WhitespaceSettingsSchema,
)

__all__ = [
    # core
    "GraphTransliterator",
    "CoverageTransliterator",
    # exceptions
    "AmbiguousTransliterationRulesException",
    "GraphTransliteratorException",
    "NoMatchingTransliterationRuleException",
    "UnrecognizableInputTokenException",
    # graphs
    "DirectedGraph",
    "VisitLoggingDirectedGraph",
    # rules
    "TransliterationRule",
    "OnMatchRule",
    "WhitespaceRules",
    # schemas
    "WhitespaceDictSettingsSchema",
    "WhitespaceSettingsSchema",
    "EasyReadingSettingsSchema",
    "OnMatchRuleSchema",
    "SettingsSchema",
    "TransliterationRuleSchema",
    "DirectedGraphSchema",
]



# -*- coding: utf-8 -*-
from collections import UserDict, UserList
from .exceptions import IncompleteGraphCoverageException
import logging

"""
Graph data structure used in Graph Transliterator.
"""
logger = logging.getLogger("graphtransliterator")


class DirectedGraph:
    """
    A very basic dictionary- and list-based directed graph. Nodes are a list of
    dictionaries of node data. Edges are nested dictionaries keyed from the
    head -> tail -> edge properties. An edge list is maintained. Can be
    exported as a dictionary.

    Attributes
    ----------
    node : `list` of `dict`
        List of node data
    edge : `dict` of {`int`: `dict` of {`int`: `dict`}}
        Mapping from head to tail of edge, holding edge data

    edge_list : `list` of `tuple` of (`int`, `int`)
        List of head and tail of each edge

    Examples
    -------
    .. jupyter-execute::

      from graphtransliterator import DirectedGraph
      DirectedGraph()

    """

    __slots__ = "edge", "node", "edge_list"

    def __init__(self, node=None, edge=None, edge_list=None):
        self.node = node if node else []
        self.edge = edge if edge else {}
        if edge:
            if not edge_list:
                edge_list = [
                    tuple([head_key, tail_key])
                    for head_key, _ in edge.items()
                    for tail_key in _.keys()
                ]
            self.edge_list = sorted(edge_list)  # sorts for string comparison
        else:
            self.edge_list = []

    def add_edge(self, head, tail, edge_data=None):
        """
        Add an edge to a graph and return its attributes as dict.

        Parameters
        ----------
        head: `int`
            Index of head of edge
        tail: `int`
            Index of tail of edge
        edge_data: `dict`, default {}
            Edge data

        Returns
        -------
        dict
            Data of created edge

        Raises
        ------
        ValueError
            Invalid ``head`` or ``tail``, or ``edge_data`` is not a dict.

        Examples
        --------
        .. jupyter-execute::

          g = DirectedGraph()
          g.add_node()

        .. jupyter-execute::

          g.add_node()

        .. jupyter-execute::

          g.add_edge(0,1, {'data_key_1': 'some edge data here'})

        .. jupyter-execute::

          g.edge

        """
        if edge_data is None:
            edge_data = {}
        if type(head) is not int:
            raise ValueError("Edge head is not an integer: %s." % head)
        if type(tail) is not int:
            raise ValueError("Edge tail is not an integer: %s." % tail)
        if head < 0 or head >= len(self.node):
            raise ValueError("Head index of edge not in graph: %s." % head)
        if tail < 0 or tail >= len(self.node):
            raise ValueError("Tail index of edge not in graph: %s." % tail)
        if type(edge_data) is not dict:
            raise ValueError("Edge data is not a dict: %s." % edge_data)

        if head not in self.edge:
            self.edge[head] = {}

        self.edge[head][tail] = edge_data

        self.edge_list.append(tuple([head, tail]))

        return self.edge[head][tail]

    def add_node(self, node_data=None):
        """Create node and return (`int`, `dict`) of node key and object.

        Parameters
        ----------
        node_data: `dict`, default {}
            Data to be stored in created node

        Returns
        -------
        `tuple` of (`int`, `dict`)
            Index of created node and its data

        Raises
        ------
        ValueError
            ``node_data`` is not a ``dict``

        Examples
        --------
        .. jupyter-execute::

          g = DirectedGraph()
          g.add_node()

        .. jupyter-execute::

          g.add_node({'datakey1': 'data value'})

        .. jupyter-execute::

          g.node

        """
        if node_data is None:
            node_data = {}
        if type(node_data) is not dict:
            raise ValueError("node_data must be a dict: %s" % node_data)
        node_key = len(self.node)
        self.node.append(node_data)
        return node_key, self.node[node_key]


class VisitLoggingList(UserList):
    """Visit logging list."""

    def __init__(self, initlist=None):
        super().__init__(initlist)
        self.visited = set()

    def __getitem__(self, i):
        self.visited.add(i)
        return self.data[i]

    def clear_visited(self):
        self.visited.clear()


class VisitLoggingDict(UserDict):
    """Visit logging dict."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.visited = set()

    def __getitem__(self, i):
        self.visited.add(i)
        return self.data[i]

    def clear_visited(self):
        self.visited.clear()


class VisitLoggingDirectedGraph(DirectedGraph):
    """A DirectedGraph that logs visits to all nodes and edges.

    Used to measure the coverage of tests for bundled transliterators."""

    def _add_visit_logging(self):
        self.node = VisitLoggingList(self.node)
        # Edges are stored  {HEAD: {TAIL: {DATA}}
        # We only care about {TAIL: DATA} for checking coverage.
        # First, convert all {TAIL: {DATA}} to VisitLoggingDict
        # (This avoids settings {HEAD: {...}} to visited
        for head in self.edge.keys():
            self.edge[head] = VisitLoggingDict(self.edge[head])
        # Second, convert {HEAD: {...}} to VisitLoggingDict
        self.edge = VisitLoggingDict(self.edge)

    def __init__(self, graph):
        """Initialize from existing graph."""
        super().__init__(edge=graph.edge, node=graph.node, edge_list=graph.edge_list)
        self._add_visit_logging()

    def clear_visited(self):
        """Clear all visited attributes on nodes and edges."""
        self.node.visited.clear()
        for head in self.edge.keys():
            self.edge[head].clear_visited()
        self.edge.clear_visited()

    def check_coverage(self, raise_exception=True):
        """Checks that all nodes and edges are visited.

        Parameters
        ----------
        raise_exception: `bool`, default
            Raise IncompleteGraphCoverageException (default, `True`)

        Returns
        -------

        Raises
        ------
        IncompleteGraphCoverageException
            Not all nodes/edges of a graph have been visited."""

        errors = False
        # Check node coverage
        missed_nodes = []
        for node_id in range(len(self.node)):
            if node_id not in self.node.visited:
                errors = True
                missed_nodes.append(node_id)
                node_data = self.node.data[node_id]  # access data so not set to visited
                logger.warning(
                    "Node {} [{}] has not been visited.".format(node_id, node_data)
                )
        # Check edge coverage
        missed_edges = []
        for head in self.edge.keys():
            for tail in self.edge.data[head].keys():  # Access data to not mark visited
                if tail not in self.edge.data[head].visited:
                    errors = True
                    missed_edges.append((head, tail))

                    logger.warning(
                        "Edge ({},{}) [{}] has not been visited.".format(
                            head, tail, self.edge.data[head].data[tail]
                        )
                    )

        # Add a comprehensives error message.
        if errors and raise_exception:
            error_msgs = []
            if missed_nodes:
                error_msgs.append("Missed nodes: " + str(missed_nodes))
            if missed_edges:
                error_msgs.append("Missed edges: " + str(missed_edges))
            raise IncompleteGraphCoverageException("; ".join(error_msgs))

        okay = not errors
        return okay



# -*- coding: utf-8 -*-
"""Tests for bundled transliterators."""

import pytest
import graphtransliterator.transliterators as transliterators


def _test_bundled_transliterator(transliterator):
    # Confirm transliteratorcan be loaded from yaml (descriptive) and json (faster)
    for method in ("yaml", "json"):

        def init_test_class(self, method=method, coverage=True):
            """Initializes test class. Sets directory and name to superclass's."""
            self._module_dir = transliterator._module_dir
            self._module_name = transliterator._module_name
            self._init_from(method, check_ambiguity=True, coverage=True)

        class_name = "Test{}From{}".format(
            type(transliterator).__name__, method.upper()
        )
        SuperClass = transliterator.__class__
        TestClass = type(class_name, (SuperClass,), {"__init__": init_test_class})
        # Confirm it is a subclass of super class
        assert issubclass(TestClass, SuperClass)
        # Check no coverage works
        assert TestClass(method="json", coverage=False)
        # Create instance of class
        test_class = TestClass(method=method)
        # Load transliteration tests (in tests/test_NAME.yaml)
        transliteration_tests = test_class.load_yaml_tests()
        # Check that bundled tests are in valid format dict {str:str}
        assert type(transliteration_tests) == dict
        for k, v in transliteration_tests.items():
            assert type(k) == str and type(v) == str
        assert len(transliteration_tests) > 0
        # Run yaml tests, calls run_tests()
        test_class.run_yaml_tests()
        # Check there is coverage of graph and onmatch rules in tests
        assert test_class.check_coverage()


def test_bundled_transliterators():
    for transliterator in transliterators.iter_transliterators():
        _test_bundled_transliterator(transliterator)


def test_duplicate_transliterator_error():
    """Test that there can be no transliterators with the same name."""
    with pytest.raises(ValueError):
        transliterators.add_transliterators(path=transliterators.__path__)


def test_iter_names():
    """Test transliterators.iter_names()."""
    assert "Example" in [_ for _ in transliterators.iter_names()]


def test_iter_transliterators():
    """Test transliterators.iter_transliterators()."""
    example = [
        _
        for _ in transliterators.iter_transliterators()
        if type(_).__name__ == "Example"
    ].pop()
    assert example.transliterate("a") == "A"


def test_transliterators_metadata():
    """Confirm that metadata of bundled transliterators matches schema."""
    for _ in transliterators.iter_transliterators():
        assert transliterators.MetadataSchema().load(_.metadata)


def test_bundled():
    """Check that Bundled exists."""
    assert transliterators.Bundled


def test_bundled_new():
    """Test new() function of Bundled."""
    assert transliterators.Example.new(method="json")
    assert transliterators.Example.new(method="yaml")


def test_no_coverage():
    """Test no coverage option."""
    x = transliterators.Example()
    x.from_JSON(coverage=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `graphtransliterator` package."""

from click.testing import CliRunner
from graphtransliterator import cli
from graphtransliterator import GraphTransliterator
from graphtransliterator import __version__ as version
from graphtransliterator.transliterators import Example
from io import StringIO
import json
import os
import yaml


def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(cli.main)
    assert result.exit_code == 0
    assert "main" in result.output
    help_result = runner.invoke(cli.main, ["--help"])
    assert help_result.exit_code == 0
    assert "Show this message and exit." in help_result.output


def test_command_line_version():
    """Test CLI version."""
    runner = CliRunner()
    version_result = runner.invoke(cli.main, ["--version"])
    assert f"{version}" in version_result.output


test_yaml = """
    tokens:
      a: [vowel]
      ' ': [whitespace]
    rules:
      a: A
      ' ': ' '
    onmatch_rules:
      - <vowel> + <vowel>: ","
    whitespace:
      consolidate: False
      default: " "
      token_class: whitespace
"""


def test_cli_transliterate_tests(tmpdir):
    """Tests transliterate command, and loading of all formats."""
    transliterator = GraphTransliterator.from_yaml(test_yaml)

    runner = CliRunner()

    # test bundled
    bundled_result = runner.invoke(
        cli.main, ["transliterate", "--from", "bundled", "Example", "a"]
    )
    assert bundled_result.exit_code == 0
    assert bundled_result.output.strip() == "A"

    # test multiple inputs with python output
    bundled_multiple_result = runner.invoke(
        cli.main, ["transliterate", "--from", "bundled", "Example", "a", "a"]
    )
    assert bundled_multiple_result.exit_code == 0
    assert bundled_multiple_result.output.strip() == str(["A", "A"])

    # test json
    bundled_multiple_json_result = runner.invoke(
        cli.main,
        ["transliterate", "--from", "bundled", "Example", "--to", "json", "a", "a"],
    )
    assert bundled_multiple_json_result.exit_code == 0
    assert bundled_multiple_json_result.output.strip() == json.dumps(["A", "A"])

    # test transliterate from JSON
    json_ = transliterator.dumps()
    json_result = runner.invoke(
        cli.main, ["transliterate", "--from", "json", json_, "a"]
    )
    assert json_result.exit_code == 0
    assert json_result.output.strip() == "A"

    # test transliterate from json file
    json_file = tmpdir.mkdir("sub").join("test.json")
    json_file.write(json_)
    json_file_result = runner.invoke(
        cli.main, ["transliterate", "--from", "json_file", json_file.strpath, "a"]
    )
    assert json_file_result.exit_code == 0
    assert json_file_result.output.strip() == "A"

    # test transliterate from yaml file
    yaml_file = tmpdir.join("test.yaml")
    yaml_file.write(test_yaml)
    yaml_file_result = runner.invoke(
        cli.main, ["transliterate", "--from", "yaml_file", yaml_file.strpath, "a"]
    )
    assert yaml_file_result.exit_code == 0
    assert yaml_file_result.output.strip() == "A"


def test_cli_generate_tests():
    """Test tests command."""
    runner = CliRunner()
    gen_tests_result = runner.invoke(
        cli.main, ["generate-tests", "--from", "bundled", "Example"]
    )
    assert gen_tests_result.exit_code == 0
    yaml_ = yaml.safe_load(StringIO(gen_tests_result.output))
    assert len(yaml_) == 5 and type(yaml_) == dict


def test_cli_dump():
    """Test `dump` command."""
    runner = CliRunner()
    dump_result = runner.invoke(cli.main, ["dump", "--from", "bundled", "Example"])
    assert dump_result.exit_code == 0
    json_ = dump_result.output
    assert GraphTransliterator.loads(json_).transliterate("a") == "A"
    # check that dump remains the same (important for version control)
    for i in range(0, 50):
        _ = runner.invoke(cli.main, ["dump", "--from", "bundled", "Example"])
        assert _.output == json_, "JSON dump varies"


def test_cli_test():
    """Test `test` command."""
    runner = CliRunner()
    test_result = runner.invoke(cli.main, ["test", "Example"])
    assert test_result.exit_code == 0
    assert test_result.output.strip() == "True"


def test_make_json():
    """Tests `make-json` command."""
    runner = CliRunner()
    # Because make-json generates a JSON file, which may differ slightly from original,
    # save and later restore original JSON file.
    orig_filename = os.path.join(Example().directory, "example.json")
    with open(orig_filename, "r") as f:
        orig_json = f.read()
    test_result = runner.invoke(cli.main, ["make-json", "Example"])
    assert test_result.exit_code == 0
    # Test regex matching
    test_result = runner.invoke(cli.main, ["make-json", "-re", "Examp"])
    assert "Made JSON of" in test_result.output
    assert test_result.exit_code == 0
    # Test regex not Matching
    test_result = runner.invoke(cli.main, ["make-json", "-re", "!"])
    assert "No bundled transliterator found " in test_result.output
    assert test_result.exit_code == 0  # no error here
    # Restore JSON file
    with open(orig_filename, "w") as f:
        f.write(orig_json)


def test_list_bundled():
    runner = CliRunner()
    test_result = runner.invoke(cli.main, ["list-bundled"])
    assert "Bundled transliterators:" in test_result.output
    assert "Example" in test_result.output



# -*- coding: utf-8 -*-

"""Tests for compression/decompression."""
import graphtransliterator
import graphtransliterator.compression as compression
from graphtransliterator import GraphTransliterator
import json
import pytest


test_config = """
tokens:
  a: [class_a]
  b: [class_b]
  c: [class_c]
  " ": [wb]
  d: []
  Aa: [contrained_rule]
rules:
  a: A
  b: B
  <class_c> <class_c> a: A(AFTER_CLASS_C_AND_CLASS_C)
  (<class_c> b) a: A(AFTER_B_AND_CLASS_C)
  (<class_c> b b) a a: AA(AFTER_BB_AND_CLASS_C)
  a <class_c>: A(BEFORE_CLASS_C)
  a b (c <class_b>): AB(BEFORE_C_AND_CLASS_B)
  c: C
  c c: C*2
  a (b b b): A(BEFORE_B_B_B)
  d (c <class_a>): D(BEFORE_C_AND_CLASS_A)
  (b b) a: A(AFTER_B_B)
  <wb> Aa: A(ONLY_A_CONSTRAINED_RULE)
  d d: "<DD>"
  d: "<D>"
  " ": " "
onmatch_rules:
  -
      <class_a> <class_b> + <class_a> <class_b>: "!"
  -
      <class_a> + <class_b>: ","
whitespace:
  default: ' '
  consolidate: True
  token_class: wb
"""


def test_compression():
    gt = GraphTransliterator.from_yaml(test_config)
    compressed_config = compression.compress_config(gt.dump())
    decompressed_config = compression.decompress_config(compressed_config)
    gt_from_decompressed = GraphTransliterator.load(decompressed_config)
    # Compare JSON dumps with sorted keys.
    assert (
        json.dumps(gt.dump(), sort_keys=True) ==
        json.dumps(gt_from_decompressed.dump(), sort_keys=True)
    )
    # Test bad compression level
    with pytest.raises(ValueError):
        gt.dump(compression_level=graphtransliterator.HIGHEST_COMPRESSION_LEVEL+1)
    # Test compression at level 0 (should likely not be called)
    assert "compressed_settings" not in compression.compress_config(
        gt.dump(), compression_level=0
    )
    # Test compression levels
    assert '"tokens": ' in gt.dumps(compression_level=0)
    assert '"compressed_settings"' in gt.dumps(compression_level=1)
    assert '"compressed_settings"' in gt.dumps(compression_level=2)
    for i in range(0, graphtransliterator.HIGHEST_COMPRESSION_LEVEL+1):
        x = gt.dumps(compression_level=i)
        y = gt.loads(x)
        assert y.transliterate("a") == "A"



# -*- coding: utf-8 -*-

"""
GraphTransliterator core classes.
"""
from .ambiguity import check_for_ambiguity
from .compression import compress_config, decompress_config
from .exceptions import (
    IncompleteOnMatchRulesCoverageException,
    IncorrectVersionException,
    NoMatchingTransliterationRuleException,
    UnrecognizableInputTokenException,
)
from .graphs import VisitLoggingDirectedGraph, VisitLoggingList
from .initialize import (
    _graph_from,
    _onmatch_rules_lookup,
    _tokenizer_pattern_from,
    _tokens_by_class_of,
    _unescape_charnames
)
from .process import _process_easyreading_settings
from .schemas import (
    DirectedGraphSchema,
    EasyReadingSettingsSchema,
    OnMatchRuleSchema,
    SettingsSchema,
    TransliterationRuleSchema,
    WhitespaceSettingsSchema,
)
from collections import deque

from graphtransliterator import __version__ as __version__
import json
import logging
from marshmallow import (
    fields, pre_load, post_load, Schema, validates_schema, ValidationError
)
import re
import yaml

logger = logging.getLogger("graphtransliterator")

DEFAULT_COMPRESSION_LEVEL = 2
HIGHEST_COMPRESSION_LEVEL = 2


class GraphTransliteratorSchema(Schema):
    """Schema for Graph Transliterator."""

    tokens = fields.Dict(
        keys=fields.Str(), values=fields.List(fields.Str()), required=True
    )
    rules = fields.Nested(TransliterationRuleSchema, many=True, required=True)
    whitespace = fields.Nested(WhitespaceSettingsSchema, many=False, required=True)
    onmatch_rules = fields.Nested(
        OnMatchRuleSchema, many=True, required=False, allow_none=True
    )
    metadata = fields.Dict(
        keys=fields.Str(), required=False  # No restriction on values
    )
    ignore_errors = fields.Bool(required=False)
    onmatch_rules_lookup = fields.Dict(required=False, allow_none=True)
    tokens_by_class = fields.Dict(
        keys=fields.Str(), values=fields.List(fields.Str), required=False
    )
    graph = fields.Nested(
        DirectedGraphSchema, many=False, allow_none=True, required=False
    )
    tokenizer_pattern = fields.Str(required=False)
    graphtransliterator_version = fields.Str(required=False)
    check_ambiguity = fields.Bool(required=False)
    # field for coverage
    coverage = fields.Bool(required=False)
    # compressed_settings = fields.Tuple(required=False)

    class Meta:
        ordered = True

    @pre_load
    def check_version_and_compression(self, data, **kwargs):
        """Raises error if serialized GraphTransliterator is from a later version."""
        version = data.get("graphtransliterator_version")
        if version and version > __version__:
            raise IncorrectVersionException
        compressed_settings = data.get("compressed_settings")
        if compressed_settings:
            data = decompress_config(compressed_settings)
            data.update(graphtransliterator_version=version)
        return data

    @post_load
    def make_GraphTransliterator(self, data, **kwargs):
        # Convert lists to sets
        for key in ("tokens", "tokens_by_class"):
            if data.get(key):  # tokens_by_class can be generated
                data[key] = {k: set(v) for k, v in data[key].items()}
        # Do not check ambiguity by default if deserializing
        data["check_ambiguity"] = kwargs.get("check_ambiguity", False)
        return GraphTransliterator(**data)

    @validates_schema
    def validate_onmatch_rules_lookup(self, data, **kwargs):
        """Check that if there are onmatch_rules_lookup there are onmatch_rules."""
        if data.get("onmatch_rules_lookup") and not data.get("onmatch_rules"):
            raise ValidationError(
                "Contains onmatch_rules_lookup but not onmatch_rules."
            )


class GraphTransliterator:
    """
    A graph-based transliteration tool that lets you convert the symbols
    of one language or script to those of another using rules that you define.

    Transliteration of tokens of an input string to an output string is
    configured by: a set of input token types with classes, pattern-matching rules
    involving sequences of tokens as well as preceding or following tokens and
    token classes, insertion rules between matches, and optional consolidation
    of whitespace. Rules are ordered by specificity.

    Note
    ----
    This constructor does not validate settings and should typically not be called
    directly. Use :meth:`from_dict` instead. For "easy reading" support, use
    :meth:`from_easyreading_dict`, :meth:`from_yaml`, or :meth:`from_yaml_file`.
    Keyword parameters used here (``ignore_errors``, ``check_ambiguity``) can be passed
    from those other constructors.

    Parameters
    ----------
    tokens : `dict` of {`str`: `set` of `str`}
        Mapping of input token types to token classes

    rules : `list` of `TransliterationRule`
        `list` of transliteration rules ordered by cost

    onmatch_rules : `list` of :class:`OnMatchRule`, or `None`
        Rules for output to be inserted between tokens
        of certain classes when a transliteration rule has been matched
        but before its production string has been added to the output

    whitespace: `WhitespaceRules`
        Rules for handling whitespace

    metadata: `dict` or `None`
        Metadata settings

    ignore_errors: `bool`, optional
        If true, transliteration errors are ignored and do not raise an
        exception. The default is false.

    check_ambiguity: `bool`, optional
        If true (default), transliteration rules are checked for ambiguity. :meth:`load()`
        and :meth:`loads` do not check ambiguity by default.

    onmatch_rules_lookup: `dict` of {`str`: dict of {`str`: `list` of `int`}}, optional`
        OnMatchRules lookup, used internally, will be generated if not present.

    tokens_by_class: `dict` of {`str`: `set` of `str`}, optional
        Tokens by class, used internally, will be generated if not present.

    graph: `DirectedGraph`, optional
        Directed graph used by Graph Transliterator, will be generated if not present.

    tokenizer_pattern: `str`, optional
        Regular expression pattern for input string tokenization, will be generated if
        not present.

    graphtransliterator_version: `str`, optional
        Version of graphtransliterator, added by `dump()` and `dumps()`.

    Example
    -------
    .. jupyter-execute::


      from graphtransliterator import GraphTransliterator, OnMatchRule, TransliterationRule, WhitespaceRules
      settings = {'tokens': {'a': {'vowel'}, ' ': {'wb'}}, 'onmatch_rules': [OnMatchRule(prev_classes=['vowel'], next_classes=['vowel'], production=',')], 'rules': [TransliterationRule(production='A', prev_classes=None, prev_tokens=None, tokens=['a'], next_tokens=None, next_classes=None, cost=0.5849625007211562), TransliterationRule(production=' ', prev_classes=None, prev_tokens=None, tokens=[' '], next_tokens=None, next_classes=None, cost=0.5849625007211562)], 'metadata': {'author': 'Author McAuthorson'}, 'whitespace': WhitespaceRules(default=' ', token_class='wb', consolidate=False)}
      gt = GraphTransliterator(**settings)
      gt.transliterate('a')


    See Also
    --------
    from_dict : Constructor from dictionary of settings
    from_easyreading_dict : Constructor from  dictionary in "easy reading" format
    from_yaml : Constructor from YAML string in "easy reading" format
    from_yaml_file : Constructor from YAML file in "easy reading" format
"""  # noqa

# ---------- initialize ----------

    def __init__(
        self,
        tokens,
        rules,
        whitespace,
        onmatch_rules=None,
        metadata=None,
        ignore_errors=False,
        check_ambiguity=True,
        onmatch_rules_lookup=None,
        tokens_by_class=None,
        graph=None,
        tokenizer_pattern=None,
        graphtransliterator_version=None,
        **kwargs
    ):
        self._tokens = tokens
        self._rules = rules
        self._tokens_by_class = tokens_by_class or _tokens_by_class_of(tokens)
        self._check_ambiguity = check_ambiguity
        if check_ambiguity:
            check_for_ambiguity(self)
        self._whitespace = whitespace

        if onmatch_rules:
            self._onmatch_rules = onmatch_rules
            if onmatch_rules_lookup:
                self._onmatch_rules_lookup = onmatch_rules_lookup
            else:
                self._onmatch_rules_lookup = _onmatch_rules_lookup(
                    tokens, onmatch_rules
                )
        else:
            self._onmatch_rules = None
            self._onmatch_rules_lookup = None

        self._metadata = metadata
        self._ignore_errors = ignore_errors

        if not tokenizer_pattern:
            tokenizer_pattern = _tokenizer_pattern_from(list(tokens.keys()))
        self._tokenizer_pattern = tokenizer_pattern
        self._tokenizer = re.compile(tokenizer_pattern, re.S)

        if not graph:
            graph = _graph_from(rules)
        self._graph = graph

        self._rule_keys = []  # last matched rules

        # When, or if, necessary, add version checking here
        if not graphtransliterator_version:
            graphtransliterator_version = __version__
        self._graphtransliterator_version = graphtransliterator_version

# ---------- class methods ----------

# ---------- private functions ----------

    def _match_constraints(self, target_edge, curr_node, token_i, tokens):
        """
        Match edge constraints.

        Called on edge before a rule. `token_i` is set to location right
        after tokens consumed.

        """
        constraints = target_edge.get("constraints")
        rules = self.rules
        if not constraints:
            return True
        for c_type, c_value in constraints.items():
            if c_type == "prev_tokens":
                num_tokens = len(rules[curr_node["rule_key"]].tokens)
                # presume for rule (a) a, with input "aa"
                # ' ', a, a, ' '  start (token_i=3)
                #             ^
                #         ^       -1 subtract num_tokens
                #      ^          - len(c_value)
                start_at = token_i
                start_at -= num_tokens
                start_at -= len(c_value)

                if not self._match_tokens(
                    start_at,
                    c_value,
                    tokens,
                    check_prev=True,
                    check_next=False,
                    by_class=False,
                ):
                    return False
            elif c_type == "next_tokens":
                # presume for rule a (a), with input "aa"
                # ' ', a, a, ' '  start (token_i=2)
                #         ^
                start_at = token_i

                if not self._match_tokens(
                    start_at,
                    c_value,
                    tokens,
                    check_prev=False,
                    check_next=True,
                    by_class=False,
                ):
                    return False

            elif c_type == "prev_classes":
                num_tokens = len(rules[curr_node["rule_key"]].tokens)
                # presume for rule (a <class_a>) a, with input "aaa"
                # ' ', a, a, a, ' '
                #                ^     start (token_i=4)
                #            ^         -num_tokens
                #         ^            -len(prev_tokens)
                #  ^                   -len(prev_classes)
                start_at = token_i
                start_at -= num_tokens
                prev_tokens = constraints.get("prev_tokens")
                if prev_tokens:
                    start_at -= len(prev_tokens)
                start_at -= len(c_value)
                if not self._match_tokens(
                    start_at,
                    c_value,
                    tokens,
                    check_prev=True,
                    check_next=False,
                    by_class=True,
                ):
                    return False

            elif c_type == "next_classes":
                # presume for rule a (a <class_a>), with input "aaa"
                # ' ', a, a, a, ' '
                #         ^          start (token_i=2)
                #            ^       + len of next_tokens (a)
                start_at = token_i
                next_tokens = constraints.get("next_tokens")
                if next_tokens:
                    start_at += len(next_tokens)
                if not self._match_tokens(
                    start_at,
                    c_value,
                    tokens,
                    check_prev=False,
                    check_next=True,
                    by_class=True,
                ):
                    return False

        return True

    def _match_tokens(
        self, start_i, c_value, tokens, check_prev=True, check_next=True, by_class=False
    ):
        """Match tokens, with boundary checks."""

        if check_prev and start_i < 0:
            return False
        if check_next and start_i + len(c_value) > len(tokens):
            return False
        for i in range(0, len(c_value)):
            if by_class:
                if not c_value[i] in self._tokens[tokens[start_i + i]]:
                    return False
            elif tokens[start_i + i] != c_value[i]:
                return False
        return True

# ---------- properties ----------

    @property
    def graph(self):
        """`DirectedGraph`: Graph used in transliteration."""
        return self._graph

    @property
    def graphtransliterator_version(self):
        """`str`: Graph Transliterator version."""
        return self._graphtransliterator_version

    @property
    def ignore_errors(self):
        """`bool`: Ignore transliteration errors setting."""
        return self._ignore_errors

    @ignore_errors.setter
    def ignore_errors(self, value):
        self._ignore_errors = value

    @property
    def last_input_tokens(self):
        """
        `list` of `str`: Last tokenization of the input string, with whitespace
        at start and end."""
        return self._input_tokens

    @property
    def last_matched_rule_tokens(self):
        """`list` of `list` of `str`: Last matched tokens for each rule."""
        return [self._rules[_].tokens for _ in self._rule_keys]

    @property
    def last_matched_rules(self):
        """
        `list` of `TransliterationRule`: Last transliteration rules matched.
        """
        return [self._rules[_] for _ in self._rule_keys]

    @property
    def metadata(self):
        """
        `dict`: Metadata of transliterator
        """
        return self._metadata

    @property
    def onmatch_rules_lookup(self):
        """
        `dict`: On Match Rules lookup
        """
        return self._onmatch_rules_lookup

    @property
    def tokenizer_pattern(self):
        """
        `str`: Tokenizer pattern from transliterator
        """
        return self._tokenizer_pattern

    @property
    def tokens_by_class(self):
        """
        `dict` of {`str`: `list` of `str`}: Tokenizer pattern from transliterator
        """
        return self._tokens_by_class

    @property
    def onmatch_rules(self):
        """`list` of :class:`OnMatchRules`: Rules for productions between matches."""
        return self._onmatch_rules

    @property
    def productions(self):
        """
        `list` of `str`: List of productions of each transliteration rule.
        """
        return [_.production for _ in self.rules]

    @property
    def rules(self):
        """
        `list` of `TransliterationRule`: Transliteration rules sorted by cost.
        """
        return self._rules

    @property
    def tokens(self):
        """
        `dict` of {`str`:`set` of `str`}: Mappings of tokens to their classes.
        """
        return self._tokens

    @property
    def whitespace(self):
        """WhiteSpaceRules: Whitespace rules."""
        return self._whitespace

# ---------- public functions ----------

    def dump(self, compression_level=0):
        """
        Dump configuration of Graph Transliterator to Python data types.

        Compression is turned off by default.

        Parameters
        ----------
        compression_level: `int`
            A value in 0 (default, no compression), 1 (compression including graph),
            and 2 (compressiong without graph)

        Returns
        -------
        OrderedDict
            GraphTransliterator configuration as a dictionary with keys:

                ``"tokens"``
                  Mappings of tokens to their classes
                  (`OrderedDict` of {str: `list` of `str`})

                ``"rules"``
                  Transliteration rules in direct format
                  (`list` of `dict` of {`str`: `str`})

                ``"whitespace"``
                  Whitespace settings
                  (`dict` of {`str`: `str`})

                ``"onmatch_rules"``
                  On match rules
                  (`list` of `OrderedDict`)

                ``"metadata"``
                  Dictionary of metadata (`dict`)

                ``"ignore_errors"``
                  Ignore errors in transliteration (`bool`)

                ``"onmatch_rules_lookup"``
                  Dictionary keyed by current token to previous token
                  containing a list of indexes of applicable :class:`OnmatchRule`
                  to try
                  (`dict` of {`str`: `dict` of {`str`: `list` of `int`}})

                ``"tokens_by_class"``
                  Tokens keyed by token class, used internally
                  (`dict` of {`str`: `list` of str})

                ``"graph"``
                  Serialization of `DirectedGraph`
                  (`dict`)

                ``"tokenizer_pattern"``
                  Regular expression for tokenizing
                  (`str`)

                ``"graphtransliterator_version"``
                  Module version of `graphtransliterator` (`str`)

        Example
        -------
        .. jupyter-execute::

          yaml_ = '''
          tokens:
            a: [vowel]
            ' ': [wb]
          rules:
            a: A
            ' ': ' '
          whitespace:
            default: " "
            consolidate: false
            token_class: wb
          onmatch_rules:
            - <vowel> + <vowel>: ','  # add a comma between vowels
          metadata:
            author: "Author McAuthorson"
          '''
          gt = GraphTransliterator.from_yaml(yaml_)
          gt.dump()


        See Also
        --------
        dumps : Dump Graph Transliterator configuration to JSON string
        load : Load Graph Transliteration from configuration in Python data types
        loads : Load Graph Transliteration from configuration as a JSON string

"""  # noqa
        if compression_level == 0:
            return GraphTransliteratorSchema().dump(
                self
            )
        elif compression_level not in range(1, HIGHEST_COMPRESSION_LEVEL+1):
            raise ValueError(
                f"Compression level must be between 0 and {HIGHEST_COMPRESSION_LEVEL}"
            )
        return {
            "graphtransliterator_version": __version__,
            "compressed_settings": compress_config(
                GraphTransliteratorSchema().dump(self),
                compression_level=compression_level
            )
        }

    def dumps(self, compression_level=2):
        """

        Parameters
        ----------
        compression_level: `int`
            A value in 0 (no compression), 1 (compression including graph),
            and 2 (default, compression without graph)
        separators: `tuple` of `str`
            Separators used by json.dumps(), default is compact

        Dump settings of Graph Transliterator to Javascript Object Notation (JSON).

        Compression is turned on by default.

        Returns
        -------
        `str`
            JSON string

        Examples
        --------
        .. jupyter-execute::

          yaml_ = '''
            tokens:
              a: [vowel]
              ' ': [wb]
            rules:
              a: A
              ' ': ' '
            whitespace:
              default: " "
              consolidate: false
              token_class: wb
            onmatch_rules:
              - <vowel> + <vowel>: ','  # add a comma between vowels
            metadata:
              author: "Author McAuthorson"
          '''
          gt = GraphTransliterator.from_yaml(yaml_)
          gt.dumps()



        See Also
        --------
        dump : Dump Graph Transliterator configuration to Python data types
        load : Load Graph Transliteration from configuration in Python data types
        loads : Load Graph Transliteration from configuration as a JSON string
        """  # noqa

        if compression_level == 0:
            return GraphTransliteratorSchema().dumps(
                self
            )
        _config = self.dump(compression_level=compression_level)
        return json.dumps(_config, separators=(',', ':'))

    def match_at(self, token_i, tokens, match_all=False):
        """
        Match best (least costly) transliteration rule at a given index in the
        input tokens and return the index to  that rule. Optionally, return all
        rules that match.

        Parameters
        ----------
        token_i : `int`
            Location in `tokens` at which to begin
        tokens : `list` of `str`
            List of tokens
        match_all : `bool`, optional
            If true, return the index of all rules matching at the given
            index. The default is false.

        Returns
        -------
        `int`, `None`, or `list` of `int`
            Index of matching transliteration rule in
            :attr:`GraphTransliterator.rules` or None. Returns a `list` of
            `int` or an empty `list` if ``match_all`` is true.

        Note
        ----
        Expects whitespaces token at beginning and end of `tokens`.

        Examples
        --------

        .. jupyter-execute::

          gt = GraphTransliterator.from_yaml('''
                  tokens:
                      a: []
                      a a: []
                      ' ': [wb]
                  rules:
                      a: <A>
                      a a: <AA>
                  whitespace:
                      default: ' '
                      consolidate: True
                      token_class: wb
          ''')
          tokens = gt.tokenize("aa")
          tokens # whitespace added to ends

        .. jupyter-execute::

          gt.match_at(1, tokens) # returns index to rule

        .. jupyter-execute::

          gt.rules[gt.match_at(1, tokens)] # actual rule

        .. jupyter-execute::

          gt.match_at(1, tokens, match_all=True) # index to rules, with match_all

        .. jupyter-execute::

          [gt.rules[_] for _ in gt.match_at(1, tokens, match_all=True)]

        """  # noqa

        graph = self._graph
        graph_node = graph.node
        graph_edge = graph.edge
        if match_all:
            matches = []
        stack = deque()

        def _append_children(node_key, token_i):
            children = None
            ordered_children = graph_node[node_key].get("ordered_children")
            if ordered_children:
                children = ordered_children.get(tokens[token_i])
                if children:
                    # reordered high to low for stack:
                    for child_key in reversed(children):

                        stack.appendleft((child_key, node_key, token_i))
                else:
                    rules_keys = ordered_children.get("__rules__")  # leafs
                    if rules_keys:
                        # There may be more than one rule, as certain rules have
                        # constraints on them.
                        # Reordered so higher cost go on stack last.
                        for rule_key in reversed(rules_keys):
                            stack.appendleft((rule_key, node_key, token_i))

        _append_children(0, token_i)  # Append all children of root node

        while stack:  # LIFO
            node_key, parent_key, token_i = stack.popleft()

            curr_node = graph_node[node_key]
            # Constraints are only on preceding edge if it is accepting
            # But edge is accessed regardless to test coverage
            incident_edge = graph_edge[parent_key][node_key]
            # Pass edge, curr_node, token index, and tokens to check constraints
            if curr_node.get("accepting") and self._match_constraints(
                incident_edge, curr_node, token_i, tokens
            ):
                if match_all:
                    matches.append(curr_node["rule_key"])
                    continue
                else:
                    return curr_node["rule_key"]
            else:
                if token_i < len(tokens) - 1:
                    token_i += 1
                _append_children(node_key, token_i)
        if match_all:
            return matches

    def pruned_of(self, productions):
        """
        Remove transliteration rules with specific output productions.

        Parameters
        ----------
        productions : `str`, or `list` of `str`
            list of productions to remove

        Returns
        -------
        graphtransliterator.GraphTransliterator
            Graph transliterator pruned of certain productions.

        Note
        ----
        Uses original initialization parameters to construct a new
        :class:`GraphTransliterator`.

        Examples
        --------
        .. jupyter-execute::

          gt = GraphTransliterator.from_yaml('''
                  tokens:
                      a: []
                      a a: []
                      ' ': [wb]
                  rules:
                      a: <A>
                      a a: <AA>
                  whitespace:
                      default: ' '
                      consolidate: True
                      token_class: wb
          ''')
          gt.rules

        .. jupyter-execute::

          gt.pruned_of('<AA>').rules

        .. jupyter-execute::

          gt.pruned_of(['<A>', '<AA>']).rules

        """  # noqa
        pruned_rules = [_ for _ in self._rules if _.production not in productions]
        return GraphTransliterator(
            self._tokens,
            pruned_rules,
            self._whitespace,
            onmatch_rules=self._onmatch_rules,
            metadata=self._metadata,
            ignore_errors=self._ignore_errors,
            check_ambiguity=self._check_ambiguity,
        )

    def tokenize(self, input):
        """
        Tokenizes an input string.

        Adds initial and trailing whitespace, which can be consolidated.

        Parameters
        ----------
        input : str
            String to tokenize

        Returns
        -------
        `list` of `str`
            List of tokens, with default whitespace token at beginning and end.

        Raises
        ------
        ValueError
            Unrecognizable input, such as a character that is not in a token

        Examples
        --------
        .. jupyter-execute::

          tokens = {'ab': ['class_ab'], ' ': ['wb']}
          whitespace = {'default': ' ', 'token_class': 'wb', 'consolidate': True}
          rules = {'ab': 'AB', ' ': '_'}
          settings = {'tokens': tokens, 'rules': rules, 'whitespace': whitespace}
          gt = GraphTransliterator.from_easyreading_dict(settings)
          gt.tokenize('ab ')

        """

        def is_whitespace(token):
            """Check if token is whitespace."""
            return self.whitespace.token_class in self.tokens[token]

        # start with a whitespace token
        tokens = [self.whitespace.default]

        prev_whitespace = True

        match_at = 0
        while match_at < len(input):
            match = self._tokenizer.match(input, match_at)
            if match:
                match_at = match.end()  # advance match_at
                token = match.group(0)
                # Could save match loc here:
                # matched_at = match.span(0)[0]
                if is_whitespace(token):
                    if prev_whitespace and self.whitespace.consolidate:
                        continue
                    else:
                        prev_whitespace = True
                else:
                    prev_whitespace = False
                tokens.append(token)
            else:
                logger.warning(
                    "Unrecognizable token %s at pos %s of %s"
                    % (input[match_at], match_at, input)
                )
                if not self.ignore_errors:
                    raise UnrecognizableInputTokenException
                else:
                    match_at += 1

        if self.whitespace.consolidate:
            while len(tokens) > 1 and is_whitespace(tokens[-1]):
                tokens.pop()

        tokens.append(self.whitespace.default)

        return tokens

    def transliterate(self, input):
        """
        Transliterate an input string into an output string.

        Parameters
        ----------
        input : `str`
            Input string to transliterate

        Returns
        -------
        `str`
            Transliteration output string

        Raises
        ------
        ValueError
            Cannot parse input

        Note
        ----
        Whitespace will be temporarily appended to start and end of input
        string.

        Example
        -------

        .. jupyter-execute::

          GraphTransliterator.from_yaml(
          '''
          tokens:
            a: []
            ' ': [wb]
          rules:
            a: A
            ' ': '_'
          whitespace:
            default: ' '
            consolidate: True
            token_class: wb
          ''').transliterate("a a")

        """
        tokens = self.tokenize(input)  # Adds initial+final whitespace
        self._input_tokens = tokens  # Tokens are saved here
        self._rule_keys = []  # Matched ule keys are saved here
        output = ""
        token_i = 1  # Adjust for initial whitespace

        while token_i < len(tokens) - 1:  # Adjust for final whitespace
            rule_key = self.match_at(token_i, tokens)
            if rule_key is None:
                logger.warning(
                    "No matching transliteration rule at token pos %s of %s"
                    % (token_i, tokens)
                )
                # No parsing rule was found at this location
                if self.ignore_errors:
                    # Move along if ignoring errors
                    token_i += 1
                    continue
                else:
                    raise NoMatchingTransliterationRuleException
            self._rule_keys.append(rule_key)
            rule = self.rules[rule_key]
            tokens_matched = rule.tokens
            if self._onmatch_rules:
                curr_match_rules = None
                prev_t = tokens[token_i - 1]
                curr_t = tokens[token_i]
                curr_t_rules = self._onmatch_rules_lookup.get(curr_t)
                if curr_t_rules:
                    curr_match_rules = curr_t_rules.get(prev_t)
                if curr_match_rules:
                    for onmatch_i in curr_match_rules:
                        onmatch = self._onmatch_rules[onmatch_i]
                        # <class_a> <class_a> + <class_b>
                        # a a b
                        #     ^
                        # ^      - len(onmatch.prev_rules)
                        if self._match_tokens(
                            token_i - len(onmatch.prev_classes),
                            onmatch.prev_classes,  # Checks last value
                            tokens,
                            check_prev=True,
                            check_next=False,
                            by_class=True,
                        ) and self._match_tokens(
                            token_i,
                            onmatch.next_classes,  # Checks first value
                            tokens,
                            check_prev=False,
                            check_next=True,
                            by_class=True,
                        ):
                            output += onmatch.production
                            break  # Only match best onmatch rule
            output += rule.production
            token_i += len(tokens_matched)
        return output

# ---------- static methods ----------

    @staticmethod
    def from_dict(dict_settings, **kwargs):
        """Generate GraphTransliterator from `dict` settings.

        Parameters
        ----------
        dict_settings : `dict`
            Dictionary of settings

        Returns
        -------
        GraphTransliterator
            Graph transliterator
        """
        settings = SettingsSchema().load(dict_settings)
        args = [
            settings["tokens"],
            settings["rules"],
            settings["whitespace"]
        ]
        kwargs = {
            "onmatch_rules": settings.get("onmatch_rules"),
            "metadata": settings.get("metadata"),
            "tokens_by_class": settings.get("tokens_by_class"),  # will be generated
            "graph": settings.get("graph"),  # will be generated
            "tokenizer_pattern": settings.get("tokenizer_pattern"),  # will be generated
            "ignore_errors": kwargs.get("ignore_errors", False),
            "check_ambiguity": kwargs.get("check_ambiguity", True),
        }
        return GraphTransliterator(*args, **kwargs)

    @staticmethod
    def from_easyreading_dict(easyreading_settings, **kwargs):
        """
        Constructs `GraphTransliterator` from a dictionary of settings in
        "easy reading" format, i.e. the loaded contents of a YAML string.

        Parameters
        ----------
        easyreading_settings : `dict`
            Settings dictionary in easy reading format with keys:

                ``"tokens"``
                  Mappings of tokens to their classes
                  (`dict` of {str: `list` of `str`})

                ``"rules"``
                  Transliteration rules in "easy reading" format
                  (`list` of `dict` of {`str`: `str`})

                ``"onmatch_rules"``
                  On match rules in "easy reading" format
                  (`dict` of {`str`: `str`}, optional)

                ``"whitespace"``
                  Whitespace definitions, including default whitespace token,
                  class of whitespace tokens, and whether or not to consolidate
                  (`dict` of {'default': `str`, 'token_class': `str`,
                  consolidate: `bool`}, optional)

                ``"metadata"``
                  Dictionary of metadata (`dict`, optional)

        Returns
        -------
        GraphTransliterator
            Graph Transliterator

        Note
        ----
        Called by :meth:`from_yaml`.

        Example
        -------
        .. jupyter-execute::

          tokens = {
              'ab': ['class_ab'],
              ' ': ['wb']
          }
          whitespace = {
              'default': ' ',
              'token_class': 'wb',
              'consolidate': True
          }
          onmatch_rules = [
              {'<class_ab> + <class_ab>': ','}
          ]
          rules = {'ab': 'AB',
                   ' ': '_'}
          settings = {'tokens': tokens,
                      'rules': rules,
                      'whitespace': whitespace,
                      'onmatch_rules': onmatch_rules}
          gt = GraphTransliterator.from_easyreading_dict(settings)
          gt.transliterate("ab abab")


        See Also
        --------
        from_yaml : Constructor from YAML string in "easy reading" format
        from_yaml_file : Constructor from YAML file in "easy reading" format
        """
        # Validate easyreading settings
        _ = EasyReadingSettingsSchema().load(easyreading_settings)
        # Convert those to regular settings
        _ = _process_easyreading_settings(_)
        # Validation of regular settings is done in from_dict
        return GraphTransliterator.from_dict(_, **kwargs)

    @staticmethod
    def from_yaml(yaml_str, charnames_escaped=True, **kwargs):
        """
        Construct GraphTransliterator from a YAML str.

        Parameters
        ----------
        yaml_str : str
            YAML mappings of tokens, rules, and (optionally) onmatch_rules
        charnames_escaped : boolean
            Unescape Unicode during YAML read (default True)

        Note
        ----
        Called by :meth:`from_yaml_file` and calls :meth:`from_easyreading_dict`.

        Example
        -------
        .. jupyter-execute::

          yaml_ = '''
          tokens:
            a: [class1]
            ' ': [wb]
          rules:
            a: A
            ' ': ' '
          whitespace:
            default: ' '
            consolidate: True
            token_class: wb
          onmatch_rules:
            - <class1> + <class1>: "+"
          '''
          gt = GraphTransliterator.from_yaml(yaml_)
          gt.transliterate("a aa")


        See Also
        --------
        from_easyreading_dict : Constructor from dictionary in "easy reading" format
        from_yaml : Constructor from YAML string in "easy reading" format
        from_yaml_file : Constructor from YAML file in "easy reading" format
        """
        if charnames_escaped:
            yaml_str = _unescape_charnames(yaml_str)

        settings = yaml.safe_load(yaml_str)

        return GraphTransliterator.from_easyreading_dict(settings, **kwargs)

    @staticmethod
    def from_yaml_file(yaml_filename, **kwargs):
        """
        Construct GraphTransliterator from YAML file.

        Parameters
        ----------
        yaml_filename : str
            Name of YAML file, containing tokens, rules, and (optionally)
            onmatch_rules

        Note
        ----
        Calls :meth:`from_yaml`.

        See Also
        --------
        from_yaml : Constructor from YAML string in "easy reading" format
        from_easyreading_dict : Constructor from dictionary in "easy reading" format
        """
        with open(yaml_filename, "r") as f:
            yaml_string = f.read()

        return GraphTransliterator.from_yaml(yaml_string, **kwargs)

    @staticmethod
    def load(settings, **kwargs):
        """Create GraphTransliterator from settings as Python data types.

        Parameters
        ----------
        settings
            GraphTransliterator configuration as a dictionary with keys:

                ``"tokens"``
                  Mappings of tokens to their classes
                  (`dict` of {str: `list` of `str`})

                ``"rules"``
                  Transliteration rules in direct format
                  (`list` of `OrderedDict` of {`str`: `str`})

                ``"whitespace"``
                  Whitespace settings
                  (`dict` of {`str`: `str`})

                ``"onmatch_rules"``
                  On match rules
                  (`list` of `OrderedDict`, optional)

                ``"metadata"``
                  Dictionary of metadata (`dict`, optional)

                ``"ignore_errors"``
                  Ignore errors. (`bool`, optional)

                ``"onmatch_rules_lookup"``
                  Dictionary keyed by current token to previous token
                  containing a list of indexes of applicable :class:`OnmatchRule`
                  to try
                  (`dict` of {`str`: `dict` of {`str`: `list` of `int`}}, optional)

                ``tokens_by_class``
                  Tokens keyed by token class, used internally
                  (`dict` of {`str`: `list` of str}, optional)

                ``graph``
                  Serialization of `DirectedGraph`
                  (`dict`, optional)

                ``"tokenizer_pattern"``
                  Regular expression for tokenizing
                  (`str`, optional)

                ``"graphtransliterator_version"``
                  Module version of `graphtransliterator` (`str`, optional)

        Returns
        -------
        GraphTransliterator
            Graph Transliterator

        Example
        -------
        .. jupyter-execute::

          from collections import OrderedDict
          settings = \
          {'tokens': {'a': ['vowel'], ' ': ['wb']},
           'rules': [OrderedDict([('production', 'A'),
                         # Can be compacted, removing None values
                         # ('prev_tokens', None),
                         ('tokens', ['a']),
                         ('next_classes', None),
                         ('next_tokens', None),
                         ('cost', 0.5849625007211562)]),
            OrderedDict([('production', ' '),
                         ('prev_classes', None),
                         ('prev_tokens', None),
                         ('tokens', [' ']),
                         ('next_classes', None),
                         ('next_tokens', None),
                         ('cost', 0.5849625007211562)])],
           'whitespace': {'default': ' ', 'token_class': 'wb', 'consolidate': False},
           'onmatch_rules': [OrderedDict([('prev_classes', ['vowel']),
                         ('next_classes', ['vowel']),
                         ('production', ',')])],
           'metadata': {'author': 'Author McAuthorson'},
           'onmatch_rules_lookup': {'a': {'a': [0]}},
           'tokens_by_class': {'vowel': ['a'], 'wb': [' ']},
           'graph': {'edge': {0: {1: {'token': 'a', 'cost': 0.5849625007211562},
              3: {'token': ' ', 'cost': 0.5849625007211562}},
             1: {2: {'cost': 0.5849625007211562}},
             3: {4: {'cost': 0.5849625007211562}}},
            'node': [{'type': 'Start', 'ordered_children': {'a': [1], ' ': [3]}},
             {'type': 'token', 'token': 'a', 'ordered_children': {'__rules__': [2]}},
             {'type': 'rule',
              'rule_key': 0,
              'accepting': True,
              'ordered_children': {}},
             {'type': 'token', 'token': ' ', 'ordered_children': {'__rules__': [4]}},
             {'type': 'rule',
              'rule_key': 1,
              'accepting': True,
              'ordered_children': {}}],
            'edge_list': [(0, 1), (1, 2), (0, 3), (3, 4)]},
           'tokenizer_pattern': '(a|\\ )',
           'graphtransliterator_version': '0.3.3'}
          gt = GraphTransliterator.load(settings)
          gt.transliterate('aa')

        .. jupyter-execute::

          # can be compacted
          settings.pop('onmatch_rules_lookup')
          GraphTransliterator.load(settings).transliterate('aa')


        See Also
        --------
        dump : Dump Graph Transliterator configuration to Python data types
        dumps : Dump Graph Transliterator configuration to JSON string
        loads : Load Graph Transliteration from configuration as a JSON string
        """  # noqa
        # combine kwargs with settings
        return GraphTransliteratorSchema().load(dict(settings, **kwargs))

    @staticmethod
    def loads(settings, **kwargs):
        """Create GraphTransliterator from JavaScript Object Notation (JSON) string.

        Parameters
        ----------
        settings
            JSON settings for GraphTransliterator

        Returns
        -------
        GraphTransliterator
            Graph Transliterator

        Example
        -------
        .. jupyter-execute::

          JSON_settings = '''{"tokens": {"a": ["vowel"], " ": ["wb"]}, "rules": [{"production": "A", "prev_classes": null, "prev_tokens": null, "tokens": ["a"], "next_classes": null, "next_tokens": null, "cost": 0.5849625007211562}, {"production": " ", "prev_classes": null, "prev_tokens": null, "tokens": [" "], "next_classes": null, "next_tokens": null, "cost": 0.5849625007211562}], "whitespace": {"default": " ", "token_class": "wb", "consolidate": false}, "onmatch_rules": [{"prev_classes": ["vowel"], "next_classes": ["vowel"], "production": ","}], "metadata": {"author": "Author McAuthorson"}, "ignore_errors": false, "onmatch_rules_lookup": {"a": {"a": [0]}}, "tokens_by_class": {"vowel": ["a"], "wb": [" "]}, "graph": {"node": [{"type": "Start", "ordered_children": {"a": [1], " ": [3]}}, {"type": "token", "token": "a", "ordered_children": {"__rules__": [2]}}, {"type": "rule", "rule_key": 0, "accepting": true, "ordered_children": {}}, {"type": "token", "token": " ", "ordered_children": {"__rules__": [4]}}, {"type": "rule", "rule_key": 1, "accepting": true, "ordered_children": {}}], "edge": {"0": {"1": {"token": "a", "cost": 0.5849625007211562}, "3": {"token": " ", "cost": 0.5849625007211562}}, "1": {"2": {"cost": 0.5849625007211562}}, "3": {"4": {"cost": 0.5849625007211562}}}, "edge_list": [[0, 1], [1, 2], [0, 3], [3, 4]]}, "tokenizer_pattern": "(a| )", "graphtransliterator_version": "1.1.1"}'''

          gt = GraphTransliterator.loads(JSON_settings)
          gt.transliterate('a')

        See Also
        --------
        dump : Dump Graph Transliterator configuration to Python data types
        dumps : Dump Graph Transliterator configuration to JSON string
        load : Load Graph Transliteration from configuration in Python data types
        """  # noqa
        # combine kwargs with settings
        _settings = dict(json.loads(settings), **kwargs)

        return GraphTransliteratorSchema().load(_settings)


class CoverageTransliterator(GraphTransliterator):
    """Subclass of GraphTransliterator that logs visits to graph and on_match rules.

    Used to confirm that tests cover the entire graph and onmatch_rules."""

    def __init__(self, *args, **kwargs):
        # Initialize from GraphTransliterator
        GraphTransliterator.__init__(self, *args, **kwargs)
        # Convert  _graph and _onmatch_rules to visit-tracking objects
        self._graph = VisitLoggingDirectedGraph(self._graph)
        self._onmatch_rules = VisitLoggingList(self._onmatch_rules)

    def clear_visited(self):
        """Clear visited flags from graph and onmatch_rules."""
        self._graph.clear_visited()
        if self._onmatch_rules:
            self._onmatch_rules.clear_visited()

    def check_onmatchrules_coverage(self, raise_exception=True):
        """Check coverage of onmatch rules."""
        errors = []
        onmatch_rules = self._onmatch_rules
        for i, onmatch_rule in enumerate(onmatch_rules.data):  # data to avoid visited
            if i not in onmatch_rules.visited:
                logger.warning(
                    "On Match Rule {} [{}] has not been visited.".format(
                        i, onmatch_rule
                    )
                )
                errors.append(i)
        if errors and raise_exception:
            error_msg = "Missed OnMatchRules: " + ",".join([str(i) for i in errors])
            raise IncompleteOnMatchRulesCoverageException(error_msg)
        return not errors

    def check_coverage(self, raise_exception=True):
        """Check coverage of graph and onmatch rules.

        First checks graph coverage, then checks onmatch rules."""

        return self._graph.check_coverage(
            raise_exception=raise_exception
        ) and self.check_onmatchrules_coverage(raise_exception=raise_exception)



# -*- coding: utf-8 -*-

from marshmallow import (
    fields,
    post_load,
    pre_dump,
    Schema,
    validate,
    ValidationError,
    validates_schema,
)

from collections import defaultdict
from .graphs import DirectedGraph
from .initialize import _onmatch_rule_of, _transliteration_rule_of, _whitespace_rules_of
from .process import RULE_RE, ONMATCH_RE
import copy


class WhitespaceDictSettingsSchema(Schema):
    """Schema for Whitespace definition as a `dict`."""

    default = fields.Str(required=True)
    token_class = fields.Str(required=True)
    consolidate = fields.Boolean(required=True)


class WhitespaceSettingsSchema(WhitespaceDictSettingsSchema):
    """Schema for Whitespace definition that loads as :class:`WhitespaceRules.`"""

    @post_load
    def make_whitespace_rules(self, data, **kwargs):
        return _whitespace_rules_of(data)


class EasyReadingSettingsSchema(Schema):
    """Schema for easy reading settings.

    Provides initial validation based on easy reading format.
    """

    tokens = fields.Dict(
        keys=fields.Str(), values=fields.List(fields.Str()), required=True
    )
    rules = fields.Dict(
        keys=fields.Str(validate=validate.Regexp(RULE_RE)),
        values=fields.Str(),
        required=True,
    )
    onmatch_rules = fields.List(
        fields.Dict(
            keys=fields.Str(validate=validate.Regexp(ONMATCH_RE)),
            values=fields.Str(),
            required=False,
        )
    )
    metadata = fields.Dict(
        keys=fields.Str(),
        # no restriction on values
        required=False,
    )
    whitespace = fields.Nested(WhitespaceDictSettingsSchema)


class TransliterationRuleSchema(Schema):
    """Schema for :class:`TransliterationRule`."""

    production = fields.Str(required=True)
    tokens = fields.List(fields.Str(), required=True)
    prev_classes = fields.List(fields.Str(), allow_none=True)
    next_classes = fields.List(fields.Str(), allow_none=True)
    prev_tokens = fields.List(fields.Str(), allow_none=True)
    next_tokens = fields.List(fields.Str(), allow_none=True)

    class Meta:
        fields = (
            "production",
            "prev_classes",
            "prev_tokens",
            "tokens",
            "next_classes",
            "next_tokens",
            "cost",
        )
        ordered = True

    @pre_dump
    def remove_nulls(self, data, **kwargs):
        """Removes None values from TransliterationRule to compress JSON."""
        # None is necessary to avoid error if production is ''
        return {k: v for k, v in data._asdict().items() if v is not None}

    @post_load
    def make_transliteration_rule(self, data, **kwargs):
        return _transliteration_rule_of(data)


class OnMatchRuleSchema(Schema):
    """Schema for :class:`OnMatchRule`."""

    prev_classes = fields.List(fields.Str())
    next_classes = fields.List(fields.Str())
    production = fields.Str()

    class Meta:
        fields = ("prev_classes", "next_classes", "production")
        ordered = True

    @post_load
    def make_onmatch_rule(self, data, **kwargs):
        return _onmatch_rule_of(data)


class SettingsSchema(Schema):
    """Schema for settings in dictionary format.

    Performs validation.
    """

    tokens = fields.Dict(
        keys=fields.Str(), values=fields.List(fields.Str()), required=True
    )
    rules = fields.Nested(TransliterationRuleSchema, many=True, required=True)
    whitespace = fields.Nested(WhitespaceSettingsSchema, many=False, required=True)
    metadata = fields.Dict(
        keys=fields.Str(), required=False  # no restriction on values
    )
    onmatch_rules = fields.Nested(OnMatchRuleSchema, many=True, required=False)

    @post_load
    def sort_rules_by_cost(self, data, **kwargs):
        """Sort rules by cost."""
        data["rules"] = sorted(data["rules"], key=lambda rule: rule.cost)
        return data

    @post_load
    def token_classes_to_sets(self, data, **kwargs):
        data["tokens"] = {k: set(v) for k, v in data["tokens"].items()}
        return data

    @validates_schema
    def validate_token_classes(self, data, **kwargs):
        errors = defaultdict(list)
        token_classes = list(set().union(*data["tokens"].values()))

        # Validate onmatch_rules and rules
        for rule_type in ("onmatch_rules", "rules"):
            for rule in data[rule_type]:
                for property in ("prev_classes", "next_classes"):
                    values = getattr(rule, property)
                    if not values:
                        continue
                    for _ in values:
                        if _ not in token_classes:
                            errors[rule_type].append(
                                'Invalid token class "{}" in {} of {}'.format(
                                    _, property, rule
                                )
                            )

        # Validate whitespace token_class
        whitespace = data["whitespace"]
        whitespace_token_class = whitespace.token_class
        if whitespace_token_class not in token_classes:
            errors["whitespace"].append(
                'Invalid token class "{}" in whitespace.'.format(
                    whitespace_token_class, whitespace
                )
            )
        if errors:
            raise ValidationError(dict(errors))

    @validates_schema
    def validate_tokens(self, data, **kwargs):
        errors = defaultdict(list)
        token_types = data["tokens"].keys()

        # Validate whitespace
        whitespace = data["whitespace"]
        default_whitespace = whitespace.default
        if default_whitespace not in token_types:
            errors["whitespace"].append(
                'Invalid default token "{}" in whitespace.'.format(default_whitespace)
            )

        # Validate rules
        rules = data["rules"]
        for rule in rules:
            for property in ("tokens", "prev_tokens", "next_tokens"):
                values = getattr(rule, property)
                if not values:
                    continue
                for _ in values:
                    if _ not in token_types:
                        errors["rules"].append(
                            'Invalid token "{}" in {} of rule {}'.format(
                                _, property, rule
                            )
                        )
        if errors:
            raise ValidationError(dict(errors))


class ConstraintSchema(Schema):
    prev_classes = fields.List(fields.Str())
    prev_tokens = fields.List(fields.Str())
    next_tokens = fields.List(fields.Str())
    next_classes = fields.List(fields.Str())


class EdgeDataSchema(Schema):
    token = fields.Str()
    cost = fields.Float()
    constraints = fields.Nested(ConstraintSchema)


class NodeDataSchema(Schema):
    token = fields.String()
    type = fields.String()
    ordered_children = fields.Dict()
    accepting = fields.Bool()
    rule_key = fields.Int()

    @pre_dump
    def strip_empty(self, data, **kwargs):
        """Remove keys with empty values, allowing zero."""
        _data = {k: v for k, v in data.items() if v or (type(v) == int and v == 0)}
        return _data


class DirectedGraphSchema(Schema):
    """ Schema for :class:`DirectedGraph`.

    Validates graph somewhat rigorously.
    """

    edge = fields.Dict(
        keys=fields.Int(),
        values=fields.Dict(keys=fields.Int, values=fields.Nested(EdgeDataSchema)),
    )
    node = fields.List(fields.Nested(NodeDataSchema))
    edge_list = fields.List(fields.Tuple((fields.Int(), fields.Int())))

    class Meta:
        fields = ("node", "edge", "edge_list")

    @post_load
    def make_graph(self, data, **kwargs):
        # Modification below alters original settings, so use a deepcopy.
        _data = copy.deepcopy(data)
        return DirectedGraph(**_data)

    @pre_dump(pass_many=True)
    def sort_edge_list(self, data, many, **kwargs):
        """Sort edge list to make dumps consistent."""
        data.edge_list = sorted(data.edge_list)
        return data
